package com.example.bump

import java.util.Base64
import javax.crypto.spec.PBEKeySpec
import javax.crypto.SecretKeyFactory
import com.macasaet.fernet.Key
import com.macasaet.fernet.Token
import com.macasaet.fernet.StringValidator
import com.macasaet.fernet.Validator
import com.maenle.bump.MainActivity
import java.math.BigInteger
import java.time.Duration
import java.time.temporal.TemporalAmount

class MessageProcessor(code: String) {
    private var sender: String
    private var password: String

    init {
        codeValid(code).let {
            val codeSplit: List<String> = code.split("-")
            sender = codeSplit.subList(0, SENDER_LENGTH).joinToString("-")
            password = codeSplit.subList(SENDER_LENGTH, codeSplit.size).joinToString("-")
        }
    }

    private fun codeValid(code: String): Boolean {
        return code.split("-").size >= KEY_LENGTH + SENDER_LENGTH
    }


    fun decrypt(messageRaw: String): String {
        val message = Message(messageRaw)
        val fernetKey: Key = message.deriveMessageKeyFromPassword(password)
        return message.validateAndDecrypt(fernetKey)
    }


    private class Message(messageRaw : String) {
        lateinit var salt : ByteArray
        var iterations: Int = 100_00
        lateinit var token : Token

        init {
            splitMessage(messageRaw)
        }

        private fun splitMessage(message : String) {
            val decodedMessage : ByteArray = Base64.getUrlDecoder().decode(message)
            val iterationsDecoded = decodedMessage.copyOfRange(SALT_LENGTH, SALT_LENGTH+ ITERATIONS_LENGTH)
            val tokenString = String(Base64.getUrlEncoder().encode(decodedMessage.copyOfRange(20, decodedMessage.size)))

            iterations = BigInteger(iterationsDecoded).toInt()
            salt = decodedMessage.copyOfRange(0, SALT_LENGTH)
            token = Token.fromString(tokenString)
        }

        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (javaClass != other?.javaClass) return false

            other as Message

            if (!salt.contentEquals(other.salt)) return false
            if (iterations != other.iterations) return false
            if (token != other.token) return false

            return true
        }

        override fun hashCode(): Int {
            var result = salt.contentHashCode()
            result = 31 * result + iterations
            result = 31 * result + token.hashCode()
            return result
        }

        fun deriveMessageKeyFromPassword(password: String): Key{
            val derivedKeyLength = 256
            val spec = PBEKeySpec(password.toCharArray(), salt, iterations, derivedKeyLength)
            val secretKeyFactory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA256")
            val key = secretKeyFactory.generateSecret(spec).encoded
            val saltedKey = Base64.getUrlEncoder().encodeToString(key)

            return Key(saltedKey)
        }

        fun validateAndDecrypt(fernetKey: Key, timeToLive: Long = 24*365*1000): String {

            val validator: Validator<String> = object : StringValidator {
                override fun getTimeToLive(): TemporalAmount {
                    return Duration.ofHours(timeToLive)
                }
            }
            return token.validateAndDecrypt(fernetKey, validator)
        }
    }

    @Suppress("unused")
    companion object {
        private val TAG = MainActivity::class.java.simpleName
        const val KEY_LENGTH = 8
        const val SENDER_LENGTH = 4
        const val SALT_LENGTH = 16
        const val ITERATIONS_LENGTH = 4
    }


}